using System;
using System.Threading.Tasks;
using FluentAssertions;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using NSubstitute;
using NSubstitute.ExceptionExtensions;
using Robware.Api.Projects.Controllers;
using Robware.Projects.Code;
using Robware.Projects.Gogs;
using Xunit;

namespace Robware.Api.Projects.Tests {
	public class CodeControllerTests {
		private class TestRepository : Repository {
			public TestRepository(string name, string url) {
				Name = name;
				Url = url;
			}
		}

		private class TestBranch : Branch {
			public TestBranch(string name, Commit commit) {
				Name = name;
				Commit = commit;
			}
		}

		private class TestCommit : Commit {
			public TestCommit(string id, string message, DateTimeOffset timestamp, string url = null) {
				Id = id;
				Message = message;
				Timestamp = timestamp;
				Url = url;
			}
		}

		[Fact]
		public async Task Repositories_WithUser_ReturnsRepositoryCollection() {
			var logger = Substitute.For<ILogger<CodeController>>();
			var api = Substitute.For<IGitApi>();
			api.GetRepositories("test").Returns(new[] { new TestRepository("test", "url") });

			var expectation = new[] { new TestRepository("test", "url") };

			var controller = new CodeController(logger, api);
			(await controller.Repositories("test")).Value.Should().BeEquivalentTo(expectation);
		}

		[Fact]
		public async Task Repositories_WithUserThatDoesntExist_Returns404() {
			var logger = Substitute.For<ILogger<CodeController>>();
			var api = Substitute.For<IGitApi>();
			api.GetRepositories("test").Throws(new UserNotFoundException("", null));

			var controller = new CodeController(logger, api);
			(await controller.Repositories("test")).Result.Should().BeOfType<NotFoundResult>();
		}

		[Fact]
		public async Task Branches_WithUserAndRepository_ReturnsBranchCollection() {
			var logger = Substitute.For<ILogger<CodeController>>();
			var api = Substitute.For<IGitApi>();
			api.GetBranches("user", "test").Returns(new[] {
				new TestBranch("master", new TestCommit("0923b554309ef562fca978c7e981b3812bc4af40", "message", new DateTimeOffset(2020, 04, 11, 14, 09, 29, TimeSpan.FromHours(1))))
			});

			var expectation = new[] {
				new TestBranch("master", new TestCommit("0923b554309ef562fca978c7e981b3812bc4af40", "message", new DateTimeOffset(2020, 04, 11, 14, 09, 29, TimeSpan.FromHours(1))))
			};

			var controller = new CodeController(logger, api);
			(await controller.Branches("user", "test")).Value.Should().BeEquivalentTo(expectation);
		}

		[Fact]
		public async Task Branches_WithUserOrRepositoryThatDoesntExist_Returns404() {
			var logger = Substitute.For<ILogger<CodeController>>();
			var api = Substitute.For<IGitApi>();
			api.GetBranches("test", "404").Throws(new RepositoryNotFoundException("", "", null));

			var controller = new CodeController(logger, api);
			(await controller.Branches("test", "404")).Result.Should().BeOfType<NotFoundResult>();
		}

		[Fact]
		public async Task Branches_WithUserAndRepositoryAndCommit_ReturnsBranchCollection() {
			var logger = Substitute.For<ILogger<CodeController>>();
			var api = Substitute.For<IGitApi>();
			api.GetCommit("user", "test", "hash")
			   .Returns(new TestCommit("0923b554309ef562fca978c7e981b3812bc4af40", "message", new DateTimeOffset(2020, 04, 11, 14, 09, 29, TimeSpan.FromHours(1))));

			var expectation = new TestCommit("0923b554309ef562fca978c7e981b3812bc4af40", "message", new DateTimeOffset(2020, 04, 11, 14, 09, 29, TimeSpan.FromHours(1)));

			var controller = new CodeController(logger, api);
			(await controller.Commit("user", "test", "hash")).Value.Should().BeEquivalentTo(expectation);
		}

		[Fact]
		public async Task Branches_WithUserOrRepositoryOrCommitThatDoesntExist_Returns404() {
			var logger = Substitute.For<ILogger<CodeController>>();
			var api = Substitute.For<IGitApi>();
			api.GetCommit("test", "404", "hash").Throws(new CommitNotFoundException("", "", "", null));

			var controller = new CodeController(logger, api);
			(await controller.Commit("test", "404", "hash")).Result.Should().BeOfType<NotFoundResult>();
		}
	}
}
