using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;
using FluentAssertions;
using Robware.Projects.Code;
using Xunit;

namespace Robware.Projects.Gogs.Tests {
	public class GogsApiTests {
		private class TestRepository : Repository {
			public TestRepository(string name, string url) {
				Name = name;
				Url = url;
			}
		}

		private class TestBranch : Branch {
			public TestBranch(string name, Commit commit) {
				Name = name;
				Commit = commit;
			}
		}

		private class TestCommit : Commit {
			public TestCommit(string id, string message, DateTimeOffset timestamp, string url = null) {
				Id = id;
				Message = message;
				Timestamp = timestamp;
				Url = url;
			}
		}

		[Fact]
		public async Task GetRepositories_WithUser_ReturnsRepositoriesCollection() {
			const string json = @"[{""id"":1,""owner"":{""id"":1,""username"":""test"",""login"":""test"",""full_name"":""Test User"",""email"":""test@test.com"",""avatar_url"":""https://test.com/avatars/1""},""name"":""Repo Name"",""full_name"":""test/repo"",""description"":"""",""private"":false,""fork"":false,""parent"":null,""empty"":false,""mirror"":false,""size"":5648384,""html_url"":""https://test.com/test/repo"",""ssh_url"":""git@test.com:test/repo.git"",""clone_url"":""https://test.com/test/repo.git"",""website"":"""",""stars_count"":0,""forks_count"":0,""watchers_count"":1,""open_issues_count"":0,""default_branch"":""master"",""created_at"":""2019-04-18T14:33:05+01:00"",""updated_at"":""2020-04-11T14:09:33+01:00"",""permissions"":{""admin"":true,""push"":true,""pull"":true}}]";
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Get)
							 .WithUrl("/api/v1/users/test/repos")
							 .WithResponse(json)
							 .Build();

			var api = new GogsApi(httpClient);
			var expectation = new[] { new TestRepository("Repo Name", "https://test.com/test/repo") };
			(await api.GetRepositories("test")).Should().BeEquivalentTo(expectation);
		}

		[Fact]
		public void GetRepositories_WithUserThatDoesntExist_ThrowsUserNotFoundException() {
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Get)
							 .WithUrl("/api/v1/users/test/repos")
							 .WithErrorStatus(HttpStatusCode.NotFound)
							 .Build();

			var api = new GogsApi(httpClient);
			api.Invoking(gogsApi => gogsApi.GetRepositories("404")).Should().Throw<UserNotFoundException>().WithMessage("Could not find user 404");
		}

		[Fact]
		public async Task GetBranches_WithUserAndRepository_ReturnsBranchCollection() {
			const string json = @"[{""name"":""master"",""commit"":{""id"":""0923b554309ef562fca978c7e981b3812bc4af40"",""message"":""message"",""url"":""Not implemented"",""author"":{""name"":""Test User"",""email"":""test@test.com"",""username"":""test""},""committer"":{""name"":""Test User"",""email"":""test@test.com"",""username"":""test""},""added"":null,""removed"":null,""modified"":null,""timestamp"":""2020-04-11T14:09:29+01:00""}}]";
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Get)
							 .WithUrl("/api/v1/repos/user/test/branches")
							 .WithResponse(json)
							 .Build();

			var api = new GogsApi(httpClient);
			var expectation = new[] {
				new TestBranch("master", new TestCommit("0923b554309ef562fca978c7e981b3812bc4af40", "message", new DateTimeOffset(2020, 04, 11, 14, 09, 29, TimeSpan.FromHours(1))))
			};
			(await api.GetBranches("user", "test")).Should().BeEquivalentTo(expectation);
		}

		[Fact]
		public void GetBranches_WithUserOrRepositoryThatDoesntExist_ThrowsRepositoryNotFoundException() {
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Get)
							 .WithUrl("/api/v1/repos/user/test/branches")
							 .WithErrorStatus(HttpStatusCode.NotFound)
							 .Build();

			var api = new GogsApi(httpClient);
			api.Invoking(gogsApi => gogsApi.GetBranches("test", "404")).Should().Throw<RepositoryNotFoundException>().WithMessage("Could not find repository 404 for user test");
		}

		[Fact]
		public async Task GetCommit_WithUserAndRepositoryAndCommitHash_ReturnsCommit() {
			const string json = @"{""url"":""https://test.com/api/v1/repos/user/repo/commits/0923b554309ef562fca978c7e981b3812bc4af40"",""sha"":""0923b554309ef562fca978c7e981b3812bc4af40"",""html_url"":""https://test.com/user/repo/commits/0923b554309ef562fca978c7e981b3812bc4af40"",""commit"":{""url"":""https://test.com/api/v1/repos/user/repo/commits/0923b554309ef562fca978c7e981b3812bc4af40"",""author"":{""name"":""Test User"",""email"":""test@test.com"",""date"":""2020-04-11T14:09:29+01:00""},""committer"":{""name"":""Test User"",""email"":""test@test.com"",""date"":""2020-04-11T14:09:29+01:00""},""message"":""message"",""tree"":{""url"":""https://test.com/api/v1/repos/user/repo/tree/0923b554309ef562fca978c7e981b3812bc4af40"",""sha"":""0923b554309ef562fca978c7e981b3812bc4af40""}},""author"":{""id"":1,""username"":""test"",""login"":""test"",""full_name"":""Test User"",""email"":""test@test.com"",""avatar_url"":""https://test.com/avatars/1""},""committer"":{""id"":1,""username"":""test"",""login"":""test"",""full_name"":""Test User"",""email"":""test@test.com"",""avatar_url"":""https://test.com/avatars/1""},""parents"":[{""url"":""https://test.com/api/v1/repos/user/repo/commits/25c320bf6be45262e49650a5a86bb71506d29617"",""sha"":""25c320bf6be45262e49650a5a86bb71506d29617""}]}";
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Get)
							 .WithUrl("/api/v1/repos/user/repo/commits/test")
							 .WithResponse(json)
							 .Build();

			var api = new GogsApi(httpClient);
			var expectation = new TestCommit("0923b554309ef562fca978c7e981b3812bc4af40", "message", new DateTimeOffset(2020, 04, 11, 14, 09, 29, TimeSpan.FromHours(1)), "https://test.com/user/repo/commit/0923b554309ef562fca978c7e981b3812bc4af40");
			(await api.GetCommit("user", "repo", "test")).Should().BeEquivalentTo(expectation);
		}

		[Fact]
		public void GetCommit_WithUserOrRepositoryOrCommitHashThatDoesntExist_ThrowsCommitNotFoundException() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Get)
			                 .WithUrl("/api/v1/repos/user/repo/commits/test")
							 .WithErrorStatus(HttpStatusCode.NotFound)
			                 .Build();

			var api = new GogsApi(httpClient);
			api.Invoking(gogsApi => gogsApi.GetCommit("test", "404", "hash")).Should().Throw<CommitNotFoundException>().WithMessage("Could not find commit hash for user test in repository 404");
		}
	}
}
